#!/usr/bin/env python

import os
import sqlite3

import logging
log = logging.getLogger()
log.addHandler(logging.NullHandler())

class StorageBase(object):
    @staticmethod
    def compress_ranges(nums):
        """Compress consequtive ranges in sequence of numbers

        E.g. [1,2,3,4,7] -> '1-4,7'
        """
        if not nums:
            return None
        nums.sort()
        buf = []
        buf.append(nums[0])
        for i in xrange(1, len(nums)):
            if nums[i-1] == nums[i]:
                pass
            elif nums[i] - nums[i-1] == 1:
                if buf[-1] is not None:
                    buf.append(None)
            else:
                if buf[-1] is None:
                    buf.append(nums[i-1])
                buf.append(nums[i])
        if buf[-1] is None:
            buf.append(nums[-1])
        output = ','.join([str(i) for i in buf])
        output = output.replace(',None,', '-')
        return output

    @staticmethod
    def expand_ranges(list_of_ranges):
        """Do the opposite of compress_ranges()

        E.g. '1-4,7' -> [1,2,3,4,7]
        """
        nums = []
        for ranges in list_of_ranges:
            for r in ranges.strip().split(','):
                if type(r) is int:
                    nums.append(r)
                elif r.isdigit():
                    nums.append(int(r))
                else:
                    try:
                        r1, r2 = r.split('-')
                        r1 = int(r1)
                        r2 = int(r2) + 1
                        nums.extend(xrange(r1, r2))
                    except ValueError as e:
                        log.error('Failed to parse chunk range "%s"' % r)
                        raise
        return nums

    @staticmethod
    def iterate_ranges(list_of_ranges):
        """Do the opposite of compress_ranges()

        E.g. '1-4,7' -> [1,2,3,4,7]
        """
        nums = []
        for ranges in list_of_ranges:
            for r in ranges.strip().split(','):
                if type(r) is int:
                    yield (r,r)
                elif r.isdigit():
                    yield (int(r), int(r))
                else:
                    try:
                        r1, r2 = r.split('-')
                        r1 = int(r1)
                        r2 = int(r2) + 1
                        yield((r1, r2))
                    except ValueError as e:
                        log.error('Failed to parse chunk range "%s"' % r)
                        raise


class SqliteStorage(StorageBase):
    """Storage abstraction for local GSB cache"""
    def __init__(self, db_path):
        self.db_path = db_path
        do_init_db = not os.path.isfile(db_path)
        log.info('Opening SQLite DB %s' % db_path)
        self.db = sqlite3.connect(db_path)
        self.dbc = self.db.cursor()
        if do_init_db:
            log.info('SQLite DB does not exist, initializing')
            self.init_db()
        self.dbc.execute('PRAGMA synchronous = 0')

    def init_db(self):
        self.dbc.execute(
        """CREATE TABLE chunk (
            chunk_number integer NOT NULL,
            timestamp timestamp DEFAULT current_timestamp,
            list_name character varying(127) NOT NULL,
            chunk_type_sub BOOLEAN NOT NULL,
            PRIMARY KEY (chunk_number, list_name, chunk_type_sub)
            )"""
        )
        self.dbc.execute(
        """CREATE TABLE full_hash (
            value BLOB NOT NULL,
            list_name character varying(127) NOT NULL,
            downloaded_at timestamp DEFAULT current_timestamp,
            expires_at timestamp without time zone NOT NULL,
            PRIMARY KEY (value, list_name)
            )"""
        )
        self.dbc.execute(
        """CREATE TABLE hash_prefix (
            value BLOB NOT NULL,
            chunk_number integer NOT NULL,
            timestamp timestamp without time zone DEFAULT current_timestamp,
            list_name character varying(127) NOT NULL,
            chunk_type_sub BOOLEAN NOT NULL,
            full_hash_expires_at timestamp NOT NULL DEFAULT current_timestamp,
            PRIMARY KEY (value, chunk_number, list_name, chunk_type_sub),
            FOREIGN KEY(chunk_number, list_name, chunk_type_sub)
                REFERENCES chunk(chunk_number, list_name, chunk_type_sub)
                ON DELETE CASCADE
            )"""
        )
        self.dbc.execute(
            """CREATE INDEX idx_hash_prefix_chunk_id ON hash_prefix (chunk_number, list_name, chunk_type_sub)"""
        )
        self.dbc.execute(
            """CREATE INDEX idx_full_hash_expires_at ON full_hash (expires_at)"""
        )
        self.db.commit()

    def chunk_exists(self, chunk):
        "Check if given chunk records already exist in the database"
        q = 'SELECT COUNT(*) FROM chunk WHERE chunk_number=? AND \
            chunk_type_sub=? AND list_name=?'
        self.dbc.execute(q, [chunk.chunk_number, (chunk.chunk_type == 'sub'), chunk.list_name])
        if self.dbc.fetchall()[0][0] > 0:
            return True
        return False

    def store_chunk(self, chunk):
        "Store chunk in the database"
        log.debug('Storing %s chunk #%s for list name %s' % (chunk.chunk_type, chunk.chunk_number, chunk.list_name))
        self.insert_chunk(chunk)
        hash_prefixes = []
        chunk_number = chunk.chunk_number
        list_name = chunk.list_name
        chunk_type_sub = chunk.chunk_type == 'sub'
        for hash_value in chunk.hashes:
            hash_prefixes.append(
                (sqlite3.Binary(hash_value), chunk_number, list_name, chunk_type_sub)
            )
        self.insert_hash_prefixes(hash_prefixes)
        self.db.commit()

    def insert_chunk(self, chunk):
        "Insert hash prefixes from the chunk to the database"
        q = 'INSERT INTO chunk (chunk_number, list_name, chunk_type_sub) \
            VALUES (?, ?, ?)'
        self.dbc.execute(q, [chunk.chunk_number, chunk.list_name, (chunk.chunk_type=='sub')])

    def insert_hash_prefixes(self, hash_prefixes):
        "Insert individual hash prefix to the database"
        q = 'INSERT INTO hash_prefix (value, chunk_number, list_name, chunk_type_sub) \
            VALUES (?, ?, ?, ?)'
        try:
            self.dbc.executemany(q, hash_prefixes)
        except sqlite3.IntegrityError as e:
            log.warn('Failed to insert chunk because of %s' % e)

    def store_full_hashes(self, hash_prefix, hashes):
        "Store hashes found for the given hash prefix"
        self.cleanup_expired_hashes()
        cache_lifetime = hashes['cache_lifetime']
        for list_name, hash_values in hashes['hashes'].items():
            for hash_value in hash_values:
                q = "INSERT INTO full_hash (value, list_name, downloaded_at, expires_at)\
                    VALUES (?, ?, current_timestamp, datetime(current_timestamp, '+%d SECONDS'))"
                self.dbc.execute(q % cache_lifetime, [sqlite3.Binary(hash_value), list_name])
        q = "UPDATE hash_prefix SET full_hash_expires_at=datetime(current_timestamp, '+%d SECONDS') \
            WHERE chunk_type_sub=? AND value=?"
        self.dbc.execute(q % cache_lifetime, [False, sqlite3.Binary(hash_prefix)])
        self.db.commit()

    def full_hash_sync_required(self, hash_prefix):
        """Check if hashes for the given hash prefix have expired

        and that prefix needs to be re-queried
        """
        q = "SELECT COUNT(*) FROM hash_prefix WHERE \
            full_hash_expires_at > current_timestamp AND chunk_type_sub=? AND value=?"
        self.dbc.execute(q, [False, sqlite3.Binary(hash_prefix)])
        c = self.dbc.fetchall()[0][0]
        return c == 0

    def lookup_full_hash(self, hash_value):
        "Query DB to see if hash is blacklisted"
        q = 'SELECT list_name FROM full_hash WHERE value=?'
        self.dbc.execute(q, [sqlite3.Binary(hash_value)])
        return [h[0] for h in self.dbc.fetchall()]

    def lookup_hash_prefix(self, hash_prefix):
        """Check if hash prefix is in the list and does not have 'sub'
        status signifying that it should be evicted from the blacklist
        """
        q = 'SELECT list_name FROM hash_prefix WHERE chunk_type_sub=? AND value=?'
        try:
            self.dbc.execute(q, [False, sqlite3.Binary(hash_prefix)])
        except sqlite3.OperationalError:
            raise RuntimeError(('Cache DB schema is incompatible with the library version. '
                                'Please remove {} and re-sync.').format(self.db_path))
        lists_add = [r[0] for r in self.dbc.fetchall()]
        if len(lists_add) == 0:
            return False
        self.dbc.execute(q, [True, sqlite3.Binary(hash_prefix)])
        lists_sub = [r[0] for r in self.dbc.fetchall()]
        if len(lists_sub) == 0:
            return True
        if set(lists_add) - set(lists_sub):
            return True
        return False

    def cleanup_expired_hashes(self):
        "Delete all hashes that behind their expiration date"
        q = 'DELETE FROM full_hash WHERE expires_at < current_timestamp'
        self.dbc.execute(q)
        self.db.commit()

    def del_chunks(self, chunk_type, list_name, chunk_numbers):
        if not chunk_numbers:
            return
        log.info('Deleting "{}" chunks {} from list {}'.format(chunk_type, repr(chunk_numbers), list_name))
        chunk_type_sub = (chunk_type == 'sub')
        for lower_boundary, upper_boundary in self.iterate_ranges(chunk_numbers):
            q = 'DELETE FROM hash_prefix WHERE chunk_type_sub=? AND list_name=? AND chunk_number>=? AND chunk_number<=?'
            self.dbc.execute(q, [chunk_type_sub, list_name, lower_boundary, upper_boundary])
            q = 'DELETE FROM chunk WHERE chunk_type_sub=? AND list_name=? AND chunk_number>=? AND chunk_number<=?'
            self.dbc.execute(q, [chunk_type_sub, list_name, lower_boundary, upper_boundary])
        self.db.commit()

    def get_existing_chunks(self):
        "Get the list of chunks that are available in the local cache"
        output = {}
        for chunk_type, chunk_type_sub in [('add', False), ('sub', True)]:
            q = """SELECT list_name, group_concat(chunk_number) FROM chunk
                WHERE chunk_type_sub=? GROUP BY list_name"""
            try:
                self.dbc.execute(q, [chunk_type_sub])
            except sqlite3.OperationalError:
                raise RuntimeError(('Cache DB schema is incompatible with the library version. '
                                    'Please remove {} and re-sync.').format(self.db_path))
            for list_name, chunks in self.dbc.fetchall():
                if not output.has_key(list_name):
                    output[list_name] = {}
                chunks = [int(c) for c in chunks.split(',')]
                output[list_name][chunk_type] = self.compress_ranges(chunks)
        return output

    def total_cleanup(self):
        "Reset local cache"
        q = 'DROP TABLE hash_prefix'
        self.dbc.execute(q)
        q = 'DROP TABLE chunk'
        self.dbc.execute(q)
        q = 'DROP TABLE full_prefix'
        self.dbc.execute(q)
        self.db.commit()
        self.init_db()



#!/usr/bin/env python

import urllib, urllib2, urlparse
import struct
import time
from StringIO import StringIO
import random
import posixpath
import re
import hashlib
import socket

from . import protobuf_pb2

import logging
log = logging.getLogger()
log.addHandler(logging.NullHandler())

class BaseProtocolClient(object):
    def __init__(self, api_key, discard_fair_use_policy=False):
        self.config = {
            "base_url": "https://safebrowsing.google.com/safebrowsing/",
            "lists": [
                "goog-malware-shavar",
                "googpub-phish-shavar",
                "goog-unwanted-shavar"
            ],
            "url_args": {
                "key": api_key,
                "appver": "0.1",
                "pver": "3.0",
                "client": "api"
            }
        }
        self.discard_fair_use_policy = discard_fair_use_policy
        self._next_call_timestamp = 0
        self._error_count = 0

    def set_next_call_timeout(self, delay):
        log.debug('Next query will be delayed %s seconds' % delay)
        self._next_call_timestamp = int(time.time()) + delay

    def fair_use_delay(self):
        "Delay server query according to Request Frequency policy"
        if self._error_count == 1:
            delay = 60
        elif self._error_count > 1:
            delay = 60 * min(480, random.randint(30, 60) * (2 ** (self._error_count - 2)))
        else:
            delay = self._next_call_timestamp - int(time.time())
        if delay > 0 and not self.discard_fair_use_policy:
            log.info('Sleeping for %s seconds' % delay)
            time.sleep(delay)

    def apiCall(self, url, payload=None):
        "Perform a call to Safe Browsing API"
        if payload is None:
            payload = ''
        request = urllib2.Request(url, data=StringIO(payload), headers={'Content-Length': len(payload)})
        try:
            response = urllib2.urlopen(request)
        except urllib2.HTTPError as e:
            self._error_count += 1
            raise
        self._error_count = 0
        return response.read()

    def mkUrl(self, service):
        "Generate Safe Browsing API URL"
        url = urllib.basejoin(self.config['base_url'], service)
        query_params = '&'.join(['%s=%s' % (k,v) for k,v in self.config['url_args'].items()])
        url = '%s?%s' % (url, query_params)
        return url


class Chunk(object):
    "Represents content of Data-response chunk content"
    def __init__(self, decoded_chunk_data, list_name):
        self.list_name = list_name
        self.hashes = []
        self.chunk_number = None
        self.chunk_type = None
        self.prefix_length = None
        self._loadChunk(decoded_chunk_data)

    def _loadChunk(self, decoded_chunk_data):
        "Decode hash prefix entries"
        hash_prefixes = []
        chunk_type = 'add'
        prefix_length = 4
        if decoded_chunk_data.chunk_type == 1:
            chunk_type = 'sub'
        if decoded_chunk_data.prefix_type == 1:
            prefix_length = 32
        hashes_str = decoded_chunk_data.hashes
        hashes_count = len(hashes_str) / prefix_length
        hashes = []
        for i in xrange(hashes_count):
            hashes.append(hashes_str[prefix_length*i:prefix_length*(i+1)])
        self.hashes = hashes
        self.chunk_number = decoded_chunk_data.chunk_number
        self.chunk_type = chunk_type
        self.prefix_length = prefix_length


class DataResponse(object):
    """Contains information on what changes need to be made

    to the local copy of hash prefixes list
    """
    def __init__(self, raw_data):
        self.del_chunks = {'add': {}, 'sub': {}}
        self.reset_required = False
        self._parseData(raw_data)

    def _parseData(self, data):
        lists_data = {}
        current_list_name = None
        for l in data:
            l = l.strip()
            if not l:
                continue
            if l.startswith('i:'):
                current_list_name = l.strip()[2:]
                lists_data[current_list_name] = []
            elif l.startswith('u:'):
                url = l[2:]
                if not url.startswith('https://'):
                    url = 'https://%s' % url
                lists_data[current_list_name].append(url)
            elif l.startswith('r:'):
                self.reset_required = True
            elif l.startswith('ad:'):
                chunk_id = l.split(':')[1]
                if current_list_name not in self.del_chunks['add']:
                    self.del_chunks['add'][current_list_name] = []
                self.del_chunks['add'][current_list_name].append(chunk_id)
            elif l.startswith('sd:'):
                chunk_id = l.split(':')[1]
                if current_list_name not in self.del_chunks['sub']:
                    self.del_chunks['sub'][current_list_name] = []
                self.del_chunks['sub'][current_list_name].append(chunk_id)
            else:
                raise RuntimeError('Response line has unexpected prefix: "%s"' % l)
        self.lists_data = lists_data

    def _unpackChunks(self, chunkDataFH):
        "Unroll data chunk containing hash prefixes"
        decoded_chunks = []
        while True:
            packed_size = chunkDataFH.read(4)
            if len(packed_size) < 4:
                break
            size = struct.unpack(">L", packed_size)[0]
            chunk_data = chunkDataFH.read(size)
            decoded_chunk = protobuf_pb2.ChunkData()
            decoded_chunk.ParseFromString(chunk_data)
            decoded_chunks.append(decoded_chunk)
        return decoded_chunks

    def _fetchChunks(self, url):
        "Download chunks of data containing hash prefixes"
        response = urllib2.urlopen(url)
        return response

    @property
    def chunks(self):
        "Generator iterating through the server respones chunk by chunk"
        for list_name, chunk_urls in self.lists_data.items():
            for chunk_url in chunk_urls:
                packed_chunks = self._fetchChunks(chunk_url)
                for chunk_data in self._unpackChunks(packed_chunks):
                    chunk = Chunk(chunk_data, list_name)
                    yield chunk


class PrefixListProtocolClient(BaseProtocolClient):
    def __init__(self, api_key, discard_fair_use_policy=False):
        super(PrefixListProtocolClient, self).__init__(api_key, discard_fair_use_policy)
        self.set_next_call_timeout(random.randint(0, 300))

    def getLists(self):
        "Get available black/white lists"
        log.info('Fetching available lists')
        url = self.mkUrl('list')
        response = self.apiCall(url)
        lists = [l.strip() for l in response.split()]
        return lists

    def _fetchData(self, existing_chunks):
        "Get references to data chunks containing hash prefixes"
        self.fair_use_delay()
        url = self.mkUrl('downloads')
        payload = []
        for l in self.config['lists']:
            list_data = existing_chunks.get(l, {})
            if not list_data:
                payload.append('%s;' % l)
                continue
            list_data_cmp = []
            if 'add' in list_data:
                list_data_cmp.append('a:%s' % list_data['add'])
            if 'sub' in list_data:
                list_data_cmp.append('s:%s' % list_data['sub'])
            payload.append('%s;%s' % (l, ':'.join(list_data_cmp)))
        payload = '\n'.join(payload) + '\n'
        response = self.apiCall(url, payload)
        return response

    def _preparseData(self, data):
        data = data.split('\n')
        next_delay = data.pop(0).strip()
        if not next_delay.startswith('n:'):
            raise RuntimeError('Expected poll interval as first line, got "%s"', next_delay)
        self.set_next_call_timeout(int(next_delay[2:]))
        return data

    def retrieveMissingChunks(self, existing_chunks={}):
        """Get list of changes from the remote server

        and return them as DataResponse object
        """
        log.info('Retrieving prefixes')
        raw_data = self._fetchData(existing_chunks)
        preparsed_data = self._preparseData(raw_data)
        d = DataResponse(preparsed_data)
        return d


class FullHashProtocolClient(BaseProtocolClient):
    def fair_use_delay(self):
        """Throttle queries according to Request Frequency policy

        https://developers.google.com/safe-browsing/developers_guide_v3#RequestFrequency
        """
        if self._error_count > 1:
            delay = min(120, 30 * (2 ** (self._error_count - 2)))
        else:
            delay = self._next_call_timestamp - int(time.time())
        if delay > 0 and self.respect_fair_use_policy:
            log.info('Sleeping for %s seconds' % delay)
            time.sleep(delay)

    def _parseHashEntry(self, hash_entry):
        "Parse full-sized hash entry"
        hashes = {}
        metadata = {}
        while True:
            if not hash_entry:
                break
            has_metadata = False
            header, hash_entry = hash_entry.split('\n', 1)
            opts = header.split(':')
            if len(opts) == 4:
                if opts[3] == 'm':
                    has_metadata = True
                else:
                    raise RuntimeError('Failed to parse full hash entry header "%s"' % header)
            list_name = opts[0]
            entry_len = int(opts[1])
            entry_count = int(opts[2])
            hash_strings = []
            metadata_strings = []
            for i in xrange(entry_count):
                hash_string = hash_entry[entry_len*i:entry_len*(i+1)]
                hash_strings.append(hash_string)
            hash_entry =  hash_entry[entry_count * entry_len:]
            if has_metadata:
                for i in xrange(entry_count):
                    next_metadata_len, hash_entry = hash_entry.split('\n', 1)
                    next_metadata_len = int(next_metadata_len)
                    metadata_str = hash_entry[:next_metadata_len]
                    metadata_strings.append(metadata_str)
                    hash_entry = hash_entry[next_metadata_len:]
            elif hash_entry:
                raise RuntimeError('Hash length does not match header declaration (no metadata)')
            hashes[list_name] = hash_strings
            metadata[list_name] = metadata_strings
        return hashes, metadata

    def getHashes(self, hash_prefixes):
        "Download and parse full-sized hash entries"
        log.info('Downloading hashes for hash prefixes %s', repr(hash_prefixes))
        url = self.mkUrl('gethash')
        prefix_len = len(hash_prefixes[0])
        hashes_len = prefix_len * len(hash_prefixes)
        p_header = '%d:%d' % (prefix_len, hashes_len)
        p_body = ''.join(hash_prefixes)
        payload = '%s\n%s' % (p_header, p_body)
        response = self.apiCall(url, payload)
        first_line, response = response.split('\n', 1)
        cache_lifetime = int(first_line.strip())
        hashes, metadata = self._parseHashEntry(response)
        return {'hashes': hashes,
                'metadata': metadata,
                'cache_lifetime': cache_lifetime,
        }


class URL(object):
    "URL representation suitable for lookup"
    def __init__(self, url):
        self.url = str(url)

    @property
    def hashes(self):
        "Hashes of all possible permutations of the URL in canonical form"
        for url_variant in self.url_permutations(self.canonical):
            url_hash = self.digest(url_variant)
            yield url_hash

    @property
    def canonical(self):
        "Convert URL to its canonical form"
        def full_unescape(u):
            uu = urllib.unquote(u)
            if uu == u:
                return uu
            else:
                return full_unescape(uu)
        def quote(s):
            safe_chars = '!"$&\'()*+,-./:;<=>?@[\\]^_`{|}~'
            return urllib.quote(s, safe=safe_chars)
        url = self.url.strip()
        url = url.replace('\n', '').replace('\r', '').replace('\t', '')
        url = url.split('#', 1)[0]
        url = quote(full_unescape(url))
        url_parts = urlparse.urlsplit(url)
        if not url_parts[0]:
            url = 'http://%s' % url
            url_parts = urlparse.urlsplit(url)
        protocol = url_parts.scheme
        host = full_unescape(url_parts.hostname)
        path = full_unescape(url_parts.path)
        query = url_parts.query
        if not query and '?' not in url:
            query = None
        if not path:
            path = '/'
        has_trailing_slash = (path[-1] == '/')
        path = posixpath.normpath(path).replace('//', '/')
        if has_trailing_slash and path[-1] != '/':
            path = path + '/'
        user = url_parts.username
        port = url_parts.port
        host = host.strip('.')
        host = re.sub(r'\.+', '.', host).lower()
        if host.isdigit():
            try:
                host = socket.inet_ntoa(struct.pack("!I", int(host)))
            except:
                pass
        if host.startswith('0x') and '.' not in host:
            try:
                host = socket.inet_ntoa(struct.pack("!I", int(host, 16)))
            except:
                pass
        quoted_path = quote(path)
        quoted_host = quote(host)
        if port is not None:
            quoted_host = '%s:%s' % (quoted_host, port)
        canonical_url = '%s://%s%s' % (protocol, quoted_host, quoted_path)
        if query is not None:
            canonical_url = '%s?%s' % (canonical_url, query)
        return canonical_url

    @staticmethod
    def url_permutations(url):
        """Try all permutations of hostname and path which can be applied
        to blacklisted URLs"""
        def url_host_permutations(host):
            if re.match(r'\d+\.\d+\.\d+\.\d+', host):
                yield host
                return
            parts = host.split('.')
            l = min(len(parts),5)
            if l > 4:
                yield host
            for i in xrange(l-1):
                yield '.'.join(parts[i-l:])
        def url_path_permutations(path):
            if path != '/':
                yield path
            query = None
            if '?' in path:
                path, query =  path.split('?', 1)
            if query is not None:
                yield path
            path_parts = path.split('/')[0:-1]
            curr_path = ''
            for i in xrange(min(4, len(path_parts))):
                curr_path = curr_path + path_parts[i] + '/'
                yield curr_path
        protocol, address_str = urllib.splittype(url)
        host, path = urllib.splithost(address_str)
        user, host = urllib.splituser(host)
        host, port = urllib.splitport(host)
        host = host.strip('/')
        for h in url_host_permutations(host):
            for p in url_path_permutations(path):
                yield '%s%s' % (h, p)

    @staticmethod
    def digest(url):
        "Hash the URL"
        return hashlib.sha256(url).digest()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf.proto',
  package='',
  serialized_pb='\n\x0eprotobuf.proto\"\xfa\x01\n\tChunkData\x12\x14\n\x0c\x63hunk_number\x18\x01 \x02(\x05\x12-\n\nchunk_type\x18\x02 \x01(\x0e\x32\x14.ChunkData.ChunkType:\x03\x41\x44\x44\x12\x35\n\x0bprefix_type\x18\x03 \x01(\x0e\x32\x15.ChunkData.PrefixType:\tPREFIX_4B\x12\x0e\n\x06hashes\x18\x04 \x01(\x0c\x12\x17\n\x0b\x61\x64\x64_numbers\x18\x05 \x03(\x05\x42\x02\x10\x01\"\x1d\n\tChunkType\x12\x07\n\x03\x41\x44\x44\x10\x00\x12\x07\n\x03SUB\x10\x01\")\n\nPrefixType\x12\r\n\tPREFIX_4B\x10\x00\x12\x0c\n\x08\x46ULL_32B\x10\x01')



_CHUNKDATA_CHUNKTYPE = _descriptor.EnumDescriptor(
  name='ChunkType',
  full_name='ChunkData.ChunkType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ADD', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUB', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=197,
  serialized_end=226,
)

_CHUNKDATA_PREFIXTYPE = _descriptor.EnumDescriptor(
  name='PrefixType',
  full_name='ChunkData.PrefixType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PREFIX_4B', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FULL_32B', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=228,
  serialized_end=269,
)


_CHUNKDATA = _descriptor.Descriptor(
  name='ChunkData',
  full_name='ChunkData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='chunk_number', full_name='ChunkData.chunk_number', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='chunk_type', full_name='ChunkData.chunk_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='prefix_type', full_name='ChunkData.prefix_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hashes', full_name='ChunkData.hashes', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='add_numbers', full_name='ChunkData.add_numbers', index=4,
      number=5, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CHUNKDATA_CHUNKTYPE,
    _CHUNKDATA_PREFIXTYPE,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=19,
  serialized_end=269,
)

_CHUNKDATA.fields_by_name['chunk_type'].enum_type = _CHUNKDATA_CHUNKTYPE
_CHUNKDATA.fields_by_name['prefix_type'].enum_type = _CHUNKDATA_PREFIXTYPE
_CHUNKDATA_CHUNKTYPE.containing_type = _CHUNKDATA;
_CHUNKDATA_PREFIXTYPE.containing_type = _CHUNKDATA;
DESCRIPTOR.message_types_by_name['ChunkData'] = _CHUNKDATA

class ChunkData(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _CHUNKDATA

  # @@protoc_insertion_point(class_scope:ChunkData)


_CHUNKDATA.fields_by_name['add_numbers'].has_options = True
_CHUNKDATA.fields_by_name['add_numbers']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), '\020\001')
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python

import logging
log = logging.getLogger()
log.addHandler(logging.NullHandler())

from .protocol import PrefixListProtocolClient, FullHashProtocolClient, URL
from .storage import SqliteStorage


class SafeBrowsingList(object):
    """Interface for Google Safe Browsing API

    supporting partial update of the local cache.
    https://developers.google.com/safe-browsing/developers_guide_v3
    """
    def __init__(self, api_key, db_path='/tmp/gsb_v3.db', discard_fair_use_policy=False):
        self.prefixListProtocolClient = PrefixListProtocolClient(api_key,
                                discard_fair_use_policy=discard_fair_use_policy)
        self.fullHashProtocolClient = FullHashProtocolClient(api_key)
        self.storage = SqliteStorage(db_path)

    def update_hash_prefix_cache(self):
        "Sync locally stored hash prefixes with remote server"
        existing_chunks = self.storage.get_existing_chunks()
        response = self.prefixListProtocolClient.retrieveMissingChunks(existing_chunks=existing_chunks)
        if response.reset_required:
            self.storage.total_cleanup()
        try:
            for chunk_type, v in response.del_chunks.items():
                for list_name, chunk_numbers in v.items():
                    self.storage.del_chunks(chunk_type, list_name, chunk_numbers)
            for chunk in response.chunks:
                if self.storage.chunk_exists(chunk):
                    log.debug('chunk #%d of type %s exists in stored list %s, skipping',
                        chunk.chunk_number, chunk.chunk_type, chunk.list_name)
                    continue
                self.storage.store_chunk(chunk)
        except:
            self.storage.db.rollback()
            raise

    def sync_full_hashes(self, hash_prefix):
        "Sync full hashes starting with hash_prefix from remote server"
        if not self.storage.full_hash_sync_required(hash_prefix):
            log.debug('Cached full hash entries are still valid for "0x%s", no sync required.', hash_prefix.encode("hex"))
            return
        full_hashes = self.fullHashProtocolClient.getHashes([hash_prefix])
        if not full_hashes:
            return
        self.storage.store_full_hashes(hash_prefix, full_hashes)

    def lookup_url(self, url):
        "Look up URL in Safe Browsing blacklists"
        url_hashes = URL(url).hashes
        for url_hash in url_hashes:
            list_name = self.lookup_hash(url_hash)
            if list_name:
                return list_name
        return None

    def lookup_hash(self, full_hash):
        """Lookup URL hash in blacklists

        Returns names of lists it was found in.
        """
        hash_prefix = full_hash[0:4]
        try:
            if self.storage.lookup_hash_prefix(hash_prefix):
                self.sync_full_hashes(hash_prefix)
                return self.storage.lookup_full_hash(full_hash)
        except:
            self.storage.db.rollback()
            raise




#!/usr/bin/env python2.7

from setuptools import setup

import sys, os

__version__ = '1.0.0'

setup(name='gglsbl',
    version=__version__,
    description="Client library for Google Safe Browsing API",
    classifiers=[
        "Operating System :: POSIX",
        "Environment :: Console",
        "Programming Language :: Python",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='safe browsing api client',
    author='Aleh Filipovich',
    author_email='aleh@vaolix.com',
    url='https://github.com/afilipovich/gglsbl',
    license='Apache2',
    packages=['gglsbl'],
    install_requires=['argparse', 'pysqlite', 'protobuf'],
    scripts=['bin/gglsbl_client.py'],
)
